import time    as _t
import spinmob as _s
import spinmob.egg as _egg
_g = _egg.gui

import traceback as _traceback
_p = _traceback.print_last

try:    import visa as _v
except: _v = None

_debug_enabled = False
def _debug(*a):
    if _debug_enabled: 
        s = []
        for x in a: s.append(str(x))
        print(', '.join(s))





class visa_api_base():
    """
    Handles the visa infrastructure common to all instrument drivers.
    
    Parameters
    ----------
    name='VISA_Alias'
        Name of the instrument, as it appears in the VISA resource manager.
    
    pyvisa_py=False
        Set to True if using pyvisa-py instead of, e.g., R&S VISA or NI-VISA.
    
    simulation=False
        Set to True to enable simulation mode.
    
    timeout=2000
        Command timeout in ms.
        
    write_sleep=0
        How many seconds to sleep after each write.
    """
    
    
    def __init__(self, name='VISA_Alias', pyvisa_py=False, simulation=False, timeout=2000, write_sleep=0.01):
        _debug('api_base.__init__()')
        # Store it
        self._write_sleep = write_sleep
        self.idn         = None
        
        # Create a resource management object
        if _v:
            if pyvisa_py: self.resource_manager = _v.ResourceManager('@py')
            else:         self.resource_manager = _v.ResourceManager()
        else:             self.resource_manager = None
        
        # If we're in simulation mode, return
        if simulation:
            self.instrument = None
            self.idn = '*** Simulation Mode ***'
            return
        
        # Try to open the instrument.
        try:
            self.instrument = self.resource_manager.open_resource(name)
            
            # Test that it's responding and is a Tektronix device.
            try:
                # Set the timeout
                self.instrument.timeout = timeout
                
                # Get the ID of the instrument
                self.idn = self.query('*IDN?').strip()
                
            except:
                print("ERROR: Instrument did not reply to IDN query. Entering simulation mode.")
                self.instrument.close()
                self.instrument = None
                self.idn = "Simulation Mode"
        
        except:
            print("ERROR: Could not open instrument. Entering simulation mode.")
            self.instrument = None
            self.idn = "Simulation Mode"
            
            # Now list all available resources
            print("Available Instruments:")
            for name in self.resource_manager.list_resources(): 
                print("  " + str(self.resource_manager.resource_info(name).alias))
        
        
    # These can be modified later to make them safe, add delays, etc.
    def command(self, message='*IDN?'):
        """
        Shortcut for coding. Runs a query() if there is a question mark, and a write() if there is not.
        """
        _debug('api_base.command('+message+')')
        
        if message.find('?') >= 0: return self.query(message)
        else:                      return self.write(message)
    
    def query(self, message='*IDN?'):
        """
        Sends the supplied message and returns the response.
        """
        _debug('api_base.query('+"'"+message+"'"+')')
        
        if self.instrument == None:
            _t.sleep(self._write_sleep)
            return
        else: 
            self.write(message)
            return self.read()
    
    def write(self, message): 
        """
        Writes the supplied message.
        """
        _debug('api_base.write('+"'"+message+"'"+')')
        
        if self.instrument == None: 
            _t.sleep(self._write_sleep)
            return
        else:
            x = self.instrument.write(message)
            _t.sleep(self._write_sleep)
            return x
        
    
    def read (self):          
        """
        Reads a message and returns it.
        """
        _debug('api_base.read()')
        
        if self.instrument == None: return
        else:                       
            s = self.instrument.read()
            _debug('  '+str(s))
            return s
    
    def read_raw(self):       
        """
        Reads a raw message (e.g. a binary stream) and returns it.
        """
        _debug('api_base.read_raw()')
        
        if self.instrument == None: return 
        else:                       
            s = self.instrument.read_raw()
            _debug('  '+ str(s))
            return s


class visa_gui_base(_g.BaseObject):
    """
    Handles the common features of our visa graphical front-ends
    
    Parameters
    ----------
    name='visa_gui'
        Make this unique for each object in a larger project. This 
        will also be the first part of the filename for the other settings files.
   
    show=True
        Whether to show the window immediately.
         
    block=False
        Whether to block the command line while showing the window.
    
    api=visa_api_base
        Class to use for the api. Should have at least the base functionality of
        visa_api_base, and take the same init arguments.
    
    timeout=2000
        Command timeout (ms)
    
    write_sleep
        How long to sleep after writing a message (sec)
    
    pyvisa_py=False
        Whether to use pyvisa_py or not.
        
    window_size=[1,1]
        Default window size.
   
    """
    def __init__(self, name='visa_gui', show=True, block=False, api=visa_api_base, timeout=2000, write_sleep=0.01, pyvisa_py=False, window_size=[1,1]):
        _debug('gui_base.__init__()')
        
        # Remember the name
        self.name = name
        self._write_sleep = write_sleep
        self._timeout     = timeout
        
        # No instrument selected yet
        self.api = None
        self._api_base = api

        # Build the GUI
        self.window    = _g.Window(name, size=window_size, autosettings_path=name+'_window.txt')
        self.grid_top  = self.window.place_object(_g.GridLayout(False))
        self.window.new_autorow()
        self.grid_bot  = self.window.place_object(_g.GridLayout(False), alignment=0)
        
        self.button_connect        = self.grid_top.place_object(_g.Button('Connect', True, False)).set_width(60)
        self.label_instrument_name = self.grid_top.place_object(_g.Label('Disconnected'), 100, 0)
        
        self.settings = self.grid_bot.place_object(_g.TreeDictionary(name+'_settings.txt', name), alignment=0)
        
        # Make sure the settings isn't the column that stretches
        self.grid_bot.set_column_stretch(1,1)
        
        # Create a resource management object
        self._pyvisa_py = pyvisa_py
        if _v:
            if pyvisa_py: self.resource_manager = _v.ResourceManager('@py')
            else:         self.resource_manager = _v.ResourceManager()
        else:             self.resource_manager = None
    
        # Get a list of resource names and a dictionary of device aliases
        # To convert from the "easy" name in the combo to the "real" name.
        names = []
        self._device_aliases = dict()
        if self.resource_manager:
            for x in self.resource_manager.list_resources():
                alias = self.resource_manager.resource_info(x).alias
                if alias == None:
                    self._device_aliases[x] = x
                    names.append(x)
                else:
                    self._device_aliases[alias] = x
                    names.append(alias)
                
        # VISA settings
        self.settings.add_parameter('VISA/Device', 0, type='list', values=['Simulation']+names)

        # Connect the signals
        self.button_connect.signal_toggled.connect(self._button_connect_clicked)
        
        # Run the base object stuff and autoload settings
        _g.BaseObject.__init__(self, autosettings_path=name)
        
        # Show the window.
        if show: self.window.show(block)

    def _button_connect_clicked(self, *a):
        """
        Connects or disconnects the VISA resource.
        """
        _debug('gui_base._button_connect_clicked', a)
        
        # If we're supposed to connect
        if self.button_connect.get_value():
            
            # Close it if it exists for some reason
            if not self.api == None: self.api.instrument.close()
            
            # Make the new one
            self.api = self._api_base(name       = self.settings['VISA/Device'], 
                                      pyvisa_py  = self._pyvisa_py,
                                      simulation = self.settings['VISA/Device']=='Simulation',
                                      timeout    = self._timeout,
                                      write_sleep= self._write_sleep)
            
            # Tell the user what scope is connected
            self.label_instrument_name.set_text(self.api.idn)
            
            # If we're in simulation mode, make it very clear
            if self.api.instrument == None:
                self.label_instrument_name.set_colors('pink' if _s.settings['dark_theme_qt'] else 'red')
                self.button_connect.set_colors(background='pink')
            else:
                self.label_instrument_name.set_style('')
                self.button_connect.set_colors(background='')
        
            # Connecting was successful!
            self._after_connect()
            
             
            
        # FAIL.
        elif not self.api == None:
            
            # Close down the instrument
            if not self.api.instrument == None: self.api.instrument.close()
            self.api = None
            self.label_instrument_name.set_text('Disconnected')
            self.button_connect.set_checked(False, block_events=True)

            # Make sure it's not still red and bold.
            self.label_instrument_name.set_style('')
            self.button_connect.set_colors(background='')
            
            # Disconnection successful!
            self._after_disconnect()


    def _after_connect(self):
        """
        Overload this function to define what happens after a successful connection.
        """
        _debug('gui_base._after_connect()')
        return

    def _after_disconnect(self):
        """
        Overload this function to define what happens after a successful disconnection.
        """
        _debug('gui_base._after_disconnect()')
        return

if __name__ == '__main__':
    self = visa_gui_base()


import numpy as _n
from scipy.special import wofz as _wofz
from scipy.special import erf  as _erf
import scipy.stats as _stats
import spinmob as _s

_ROOT2 = 2.0**0.5 # Code speedup

def erfcx(x):
    """
    Scaled complementary error function.
    """
    return _n.exp(x**2)*(1-_erf(x))

def em_gaussian(x, sigma=1, tau=1):
    """
    Returns an exponentially modified Gaussian (a convolution of an exponential
    cutoff at x=0 and Gaussian) having standard deviation sigma and exponential 
    decay length tau. This function is normalized to have unity area.
    
    Parameters
    ----------
    x:
        Distance from the center of the peak.
    sigma:
        Standard deviation of Gaussian ~ exp(-x**2/(2*sigma**2))
    tau:
        Length scale of exponential ~ exp(x/tau). Positive tau skews the peak 
        to higher values and negative tau skews to lower values.
    """
    t = abs(tau)
    s = sigma
    
    if tau >= 0: return 0.5/t*_n.exp(-0.5*( x/s)**2)*erfcx((s/t - x/s)*0.5**0.5)
    else:        return 0.5/t*_n.exp(-0.5*(-x/s)**2)*erfcx((s/t + x/s)*0.5**0.5)

def voigt(x, sigma=1, gamma=1):
    """
    Returns a Voigt function (a convolution of a Lorentzian and Gaussian) 
    centered at x=0 with Gaussian standard deviation sigma and Lorentzian 
    half-width gamma. The function is normalized to have unity area.
    
    Parameters
    ----------
    x:
        Distance from center of peak.
    sigma = 1:
        Standard deviation of Gaussian ~ exp(-x**2/(2*sigma**2)) 
    gamma = 1:
        Halfwidth of Lorentzian ~ 1/(1+x**2/gamma**2)
    """
    return _n.real(_wofz((x + 1j*gamma)/sigma/_ROOT2)) / sigma / (2*_n.pi)**0.5

def reduced_chi2(x, dof):
    """
    Returns the reduced chi^2 probability density function for the specified
    degrees of freedom (dof).
    
    Parameters
    ----------
    x
        Value of reduced chi^2.
    
    dof
        Degrees of freedom.
    """
    return dof*_stats.chi2.pdf(x*dof,dof)




import sys as _sys
import os  as _os
import spinmob



# Get the version
try: exec(spinmob.fun.read_lines(_os.path.join(__path__[0],'setup.py'))[0])
except: __version__ = 'unknown'

# Windows ADALM2000 Drivers
if _sys.platform in ['win32']:
    try:    import libm2k as _m2k
    except:
        _m2k = None
        spinmob._warn('To use an ADALM2000 on Windows, you need to install libm2k with python bindings.')


# OSX ADALM2000 Drivers
elif _sys.platform in ['darwin']:
    spinmob._warn('No ADALM2000 on OSX yet.')
    _m2k = None

# Linux ADALM2000 Drivers
else:
    _sys.path.append(_os.path.join(__path__[0], 'libm2k', 'linux'))
    try: import libm2k as _m2k
    except:
        _m2k = None
        spinmob._warn('To use an ADALM2000 on Linux, you need to install libiio v0.21, and libm2k v0.2.1.')


# Test for VISA
try:    import visa as _v
except:
    spinmob._warn('Visa driver and / or pyvisa not installed. On Windows, consider Rhode & Schwartz VISA or NI-VISA, then pip install pyvisa. On Linux, pip install pyvisa and pyvisa-py')
    _v = None

from . import visa_tools
from . import instruments
from . import data
from . import functions
from . import playground


import os        as _os
import numpy     as _n
import spinmob   as _s
import mcphysics as _m
import unittest  as _ut
import shutil    as _sh

# Globals for monkeywork on the command line
a = b = c = d = e = x = None

if _os.path.exists('egg_settings'): _sh.rmtree('egg_settings')

# Path to the data folder
data_path = _os.path.join(_os.path.dirname(_m.__file__), 'tests', 'data')
def path(filename):
    """
    Assembles a path to a particular file.
    """
    return _os.path.join(data_path, filename)


class Test_errthing(_ut.TestCase):
    """
    Test class for mcphysics library.
    """


    def setUp(self):
        return

    def tearDown(self):
        return

    def test_data(self):

        d = _m.data.load_chn(path('signal.Chn'))
        self.assertEqual(len(d), 2)
        self.assertEqual(len(d[1]), 1024)

        ds = _m.data.load_chns([path('signal.Chn'), path('background.Chn')])
        self.assertEqual(len(ds), 2)
        self.assertEqual(len(ds[1]), 2)

        _s.pylab.figure(1)
        _m.data.plot_chn_files(paths=[path('signal.Chn'), path('background.Chn')])

        image = _m.data.load_image(path('image.jpg'))
        self.assertEqual(image.shape, (612, 816, 3))

        images = _m.data.load_images([path('image.jpg')])
        self.assertEqual(_n.shape(images), (1,612,816,3))

        ds = _m.data.convert_chn_to_csv([path('signal.Chn'),path('background.Chn')])
        

    def test_functions(self):

        _s.pylab.figure(2)
        _s.plot.xy.function(['em_gaussian(x,1,2)', 'voigt(x,2,1)', 'erfcx(x)', 'reduced_chi2(x,10)'],
                             1e-6,5,1000,g=_m.functions.__dict__)

    def test_instruments(self):
        global a, b, c

        a = _m.instruments.adalm2000(block=True)
        b = _m.instruments.sillyscope(block=True)
        c = _m.instruments.keithley_dmm(block=True)


    def test_playground(self):
        global b

        b = _m.playground.fitting_statistics_demo(block=True)

        print()
        _m.playground.plot_and_integrate_reduced_chi2()

if __name__ == "__main__":
    _ut.main()



 # This file is part of the Macrospyn distribution 
 # (https://github.com/Spinmob/macrospyn).
 # Copyright (c) 2002-2020 Jack Childress (Sankey).
 # 
 # This program is free software: you can redistribute it and/or modify  
 # it under the terms of the GNU General Public License as published by  
 # the Free Software Foundation, version 3.
 # 
 # This program is distributed in the hope that it will be useful, but 
 # WITHOUT ANY WARRANTY; without even the implied warranty of 
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU 
 # General Public License for more details.
 # 
 # You should have received a copy of the GNU General Public License 
 # along with this program. If not, see <http://www.gnu.org/licenses/>.

import numpy       as _n
import scipy.stats as _stats

# For embedding matplotlib figures
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg    as _canvas
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as _navbar
from matplotlib.figure                  import Figure               as _figure

import mcphysics   as _m 
import spinmob     as _s
import spinmob.egg as _egg
_g = _egg.gui

import traceback as _traceback
_p = _traceback.print_last



def plot_and_integrate_reduced_chi2(dof=10, xmin=1e-6, xmax=5, steps=1e5):
    """
    Plots the reduced chi^2 density function, and then numerically integrates it.
    
    Parameters
    ----------
    dof=10
        Degrees of freedom.
    xmin, xmax, steps
        Plot range from xmin to xmax with the specified steps. This will affect
        the validity of the numerical integral.
    """
    
    _s.pylab.figure(100)
    a1 = _s.pylab.subplot(211)
    a2 = _s.pylab.subplot(212, sharex=a1)
    
    _s.plot.xy.function('f(x,dof)', xmin, xmax, steps, g=dict(f=_m.functions.reduced_chi2, dof=dof), 
                        axes=a1, ylabel='$P(\chi^2_r)$', xlabel='$\chi^2_r$', tall=True)
    _s.tweaks.integrate_shown_data(output_axes=a2, tall=True)
    _s.tweaks.ubertidy(window_size=[800,950])





class fitting_statistics_demo():
    
    """
    Graphical interface for generating fake data, fitting, and collecting 
    fit statistics.
    
    Parameters
    ----------
    block=True
        Whether to block the command line when the window is first shown.
    """
    
    
    def __init__(self, block=False):
        
        self._build_gui(block)

    def _build_gui(self, block_command_line=False):
        """
        Builds the GUI for taking fake data.
        """
        # Make a window with a left grid for settings and controls, and
        # a right grid for visualization.
        self.window = _g.Window('Fake Data Taker', size=[1000,700], autosettings_path='window.cfg')
        self.window.event_close = self.event_close
        self.grid_controls = self.window.place_object(_g.GridLayout(False))
        self.grid_plotting = self.window.place_object(_g.GridLayout(False), alignment=0)
        
        # Add the acquire button & connect the signal
        self.button_acquire = self.grid_controls.place_object(_g.Button('Acquire'),    alignment=0).set_width(70)
        self.button_fit     = self.grid_controls.place_object(_g.Button('Fit')    ,    alignment=0).set_width(55)
        self.button_loop    = self.grid_controls.place_object(_g.Button('Loop', True), alignment=0).set_width(55)
        self.button_clear   = self.grid_controls.place_object(_g.Button('Clear'),      alignment=0).set_width(55)
        self.button_acquire.signal_clicked.connect(self.button_acquire_clicked)
        self.button_fit    .signal_clicked.connect(self.button_fit_clicked)
        self.button_loop   .signal_clicked.connect(self.button_loop_clicked)
        self.button_clear  .signal_clicked.connect(self.button_clear_clicked)
        
        # Create an populate the settings tree
        self.grid_controls.new_autorow()
        self.tree_settings  = self.grid_controls.place_object(_g.TreeDictionary(), column_span=4)
        
        self.tree_settings.add_parameter('Acquire/reality', '1.7*x+1.2')
        self.tree_settings.add_parameter('Acquire/x_noise',        0)
        self.tree_settings.add_parameter('Acquire/y_noise',      1.3)
        
        self.tree_settings.add_parameter('Acquire/xmin',    0)
        self.tree_settings.add_parameter('Acquire/xmax',   10)
        self.tree_settings.add_parameter('Acquire/steps', 100, dec=True)
        
        self.tree_settings.add_parameter('Fit/function',   'a*x+b')
        self.tree_settings.add_parameter('Fit/parameters', 'a=0,b=0')
        self.tree_settings.add_parameter('Fit/assumed_ey', 1.3)
        
        self.tree_settings.add_parameter('Stats/bins',        14)
        self.tree_settings.add_parameter('Stats/versus_x',    'a')
        self.tree_settings.add_parameter('Stats/versus_y',    'b')
        self.tree_settings.add_parameter('Stats/plot_theory', False)
        
        # Add the tabs and plotter to the other grid
        self.tabs_plotting = self.grid_plotting.place_object(_g.TabArea('tabs_plotting.cfg'), alignment=0)
        
        # Tab for raw data
        self.tab_raw  = self.tabs_plotting.add_tab('Raw Data')
        self.plot_raw = self.tab_raw.place_object(
                _g.DataboxPlot(autosettings_path='plot_raw.cfg', autoscript=1), 
                alignment=0)
        self.plot_raw.autoscript_custom = self._autoscript_raw
        
        # Tab for fit
        self.tab_fit    = self.tabs_plotting.add_tab('Fit')
        
        self.figure_fit = _figure()
        self.canvas_fit = _canvas(self.figure_fit)
        self.navbar_fit = _navbar(self.canvas_fit, self.window._widget)
        
        self.tab_fit.place_object(self.navbar_fit, alignment=0)
        self.tab_fit.new_autorow()
        self.tab_fit.place_object(self.canvas_fit, alignment=0)

        # Fitter object linked to this figure canvas
        self.fitter = _s.data.fitter()
        self.fitter.set(autoplot=False)
        self.fitter.figures = [self.figure_fit]

        # Tab for running total of fit parameters
        self.tab_parameters  = self.tabs_plotting.add_tab('Fit Parameters')
        self.plot_parameters = self.tab_parameters.place_object(
                _g.DataboxPlot(autosettings_path='plot_parameters.cfg'),
                alignment=0)
        # Give it a handle on the fitter for the script
        self.plot_parameters.fitter = self.fitter
        
        # Tab for histograms
        self.tab_stats    = self.tabs_plotting.add_tab('Histograms')
        
        self.figure_stats = _figure()
        self.canvas_stats = _canvas(self.figure_stats)
        self.navbar_stats = _navbar(self.canvas_stats, self.window._widget)
        
        self.tab_stats.place_object(self.navbar_stats, alignment=0)
        self.tab_stats.new_autorow()
        self.tab_stats.place_object(self.canvas_stats, alignment=0)
        
        # Changing tabs can update plots
        self.tabs_plotting.signal_switched.connect(self.tabs_plotting_switched)
        
        # Set up the autosave & load.
        self.tree_settings.connect_any_signal_changed(self.tree_settings.autosave)
        self.tree_settings.connect_any_signal_changed(self.update_all_plots)
        self.tree_settings.load()
        
        # Show the window
        self.window.show(block_command_line)
    
    def _autoscript_raw(self):
        """
        Returns a nice custom autoscript for plotting the raw data.
        """
        return "x = [ d[0] ]\ny = [ d[1] ]\n\nxlabels = 'x'\nylabels = ['y']"
    
    def tabs_plotting_switched(self, *a):
        """
        Someone switched a tab!
        """
        if   a[0]==1: self.update_fit_plot()
        elif a[0]==3: self.update_histograms_plot()
    
    def button_acquire_clicked(self, *a):
        """
        Acquires fake data and dumps it with the header into the plotter.
        """
        
        # Dump the header info
        self.tree_settings.send_to_databox_header(self.plot_raw)
        
        # Generate the data
        x = _n.linspace(self.tree_settings['Acquire/xmin'], 
                        self.tree_settings['Acquire/xmax'],
                        self.tree_settings['Acquire/steps'])
        
        
        d = _s.fun.generate_fake_data(self.tree_settings['Acquire/reality'], x,
                                      self.tree_settings['Acquire/y_noise'],
                                      self.tree_settings['Acquire/x_noise'])
        
        # Dump it to the plotter and plot
        self.plot_raw.copy_columns(d)
        
        # Plot it.
        self.plot_raw.plot()
        
        # Autosave if checked
        self.plot_raw.autosave()
    
    def button_fit_clicked(self,*a):
        """
        Assuming there is data, run the fit!
        """
        # Set the functions
        self.fitter.set_functions(self.tree_settings['Fit/function'],
                                  self.tree_settings['Fit/parameters'])
        
        # Set the data
        self.fitter.set_data(self.plot_raw[0], self.plot_raw[1], 
                             self.tree_settings['Fit/assumed_ey'])
        
        # Fit!
        self.fitter.fit()
        
        # Draw
        self.figure_fit.canvas.draw()
        self.window.process_events()
        
        # Now append the fit results to the next tab's plotter
        ps  = self.fitter.results.params
        x2  = self.fitter.get_reduced_chi_squared()
        dof = self.fitter.get_degrees_of_freedom()
        
        ckeys = ['reduced_chi2', 'DOF']
        row   = [x2,dof]
        for pname in ps:
            
            # Append the fit parameter
            ckeys.append(pname)
            row  .append(ps[pname].value)

            # Append the fit error
            ckeys.append(pname+'_error')
            row  .append(ps[pname].stderr)
        
        # If the parameters haven't changed, just append the data
        self.plot_parameters.append_row(row, ckeys=ckeys)
        
        # If this is the first row, set up the histograms
        if len(self.plot_parameters[0]) == 1:
            
            # PARAMETERS: Send the settings to the header
            self.tree_settings.send_to_databox_header(self.plot_parameters)
            
            # Generate a plot script
            s = 'x = [None]\ny = [d[0],d[1]'
            for n in range(len(ps)):
                s = s+',d['+str(2*n+2)+']'
            
            s = s+']\n\nxlabels = "Iteration"\nylabels = [ d.ckeys[0], d.ckeys[1]'
            for n in range(len(ps)):
                s = s+',d.ckeys['+str(2*n+2)+']'
            s = s+']'
            
            # Set to manual script and update the text
            self.plot_parameters.combo_autoscript.set_value(0, block_events=True)
            self.plot_parameters.script.set_text(s)
            
            # HISTOGRAMS: Clear the figure and set up the histogram axes
            self.axes_histograms = []
            self.figure_stats.clear()
        
            # Calculate how many rows of plots are needed
            rows = int(_n.ceil(len(ps)*0.5)+1)
        
            # Reduced chi^2 histogram
            self.axes_histograms.append(self.figure_stats.add_subplot(rows, 2, 1))
            self.axes_histograms.append(self.figure_stats.add_subplot(rows, 2, 2))
        
            # Parameter histograms
            for n in range(len(ps)):
                self.axes_histograms.append(self.figure_stats.add_subplot(rows, 2, n+3))
                    
        # Update the parameters plot!
        self.plot_parameters.plot()
        
        # If we're on the fit or stats tab (these are slow to plot)
        if self.tabs_plotting.get_current_tab()==1: self.update_fit_plot()
        if self.tabs_plotting.get_current_tab()==3: self.update_histograms_plot()
        
    def button_loop_clicked(self, value):
        """
        When someone clicks the "loop" button. 
        """
        # If it's enabled, start the loop
        if not value: return
    
        # Run the loop
        while self.button_loop.is_checked():
            
            # Acquire data and fit
            self.button_acquire_clicked(True)
            self.window.process_events()
            self.button_fit_clicked(True)
            self.window.process_events()
    
    def button_clear_clicked(self, *a):
        """
        Someone clears the data.
        """
        self.plot_parameters.clear()
        self.update_all_plots()
    
    def update_fit_plot(self):
        """
        Update the fit plot.
        """
        if not self.tabs_plotting.get_current_tab()==1: return
        
        self.fitter.plot()
        self.window.process_events()
        self.figure_fit.canvas.draw()
        self.window.process_events()
    
    def update_histograms_plot(self):
        """
        Update the histogram plots (actually perform the histogram and plot).
        """
        # Don't bother if we're not looking.
        if not self.tabs_plotting.get_current_tab()==3: return
        
        if len(self.plot_parameters) and len(self.axes_histograms):
            
            # Update the chi^2 histogram histograms
            self.axes_histograms[0].clear()
            N,B,c = self.axes_histograms[0].hist(self.plot_parameters[0], self.tree_settings['Stats/bins'], label='$\chi^2_{reduced}$')
            x = (B[1:]+B[:-1])*0.5
            
            # Include the error bars
            self.axes_histograms[0].errorbar(x, N, _n.sqrt(N), ls='', marker='+')
            
            # Tidy up
            self.axes_histograms[0].set_xlabel('$\chi^2_{reduced}$')
            self.axes_histograms[0].set_ylabel('Counts')
            
            # Plot the expected distribution.
            if self.tree_settings['Stats/plot_theory']:
                
                x2  = _n.linspace(min(0.5*(B[1]-B[0]),0.02), max(1.5,max(self.plot_parameters[0])), 400)
                dof = self.plot_parameters[1][-1]
                pdf = len(self.plot_parameters[1]) * dof * _stats.chi2.pdf(x2*dof,dof) * (B[1]-B[0])                
                self.axes_histograms[0].plot(x2,pdf,label='Expected ('+str(dof)+ 'DOF)')
                self.axes_histograms[0].legend()
            
            # Include zero, to give a sense of scale.
            self.axes_histograms[0].set_xlim(0,max(1.5,max(self.plot_parameters[0]))*1.05)
            
            
            # Plot the correlations
            self.axes_histograms[1].clear()
            self.axes_histograms[1].plot(self.plot_parameters[self.tree_settings['Stats/versus_x']],
                                         self.plot_parameters[self.tree_settings['Stats/versus_y']],
                                         label=self.tree_settings['Stats/versus_y']+' vs '+self.tree_settings['Stats/versus_x'],
                                         linestyle='', marker='o', alpha=0.3)
            self.axes_histograms[1].set_xlabel(self.tree_settings['Stats/versus_x'])
            self.axes_histograms[1].set_ylabel(self.tree_settings['Stats/versus_y'])
            self.axes_histograms[1].legend()
            
            # Now plot the distributions of the other fit parameters.
            for n in range(len(self.fitter.p_fit)):
                
                # Plot the histogram
                self.axes_histograms[n+2].clear()
                N,B,c = self.axes_histograms[n+2].hist(self.plot_parameters[2*n+2], self.tree_settings['Stats/bins'], label=self.fitter.get_parameter_names()[n])
                x = (B[1:]+B[:-1])*0.5
                
                # Include the error bars
                self.axes_histograms[n+2].errorbar(x, N, _n.sqrt(N), ls='', marker='+')
            
                # Tidy up
                self.axes_histograms[n+2].set_xlabel(self.fitter.get_parameter_names()[n])
                self.axes_histograms[n+2].set_ylabel('Counts')
                
                # Plot the expected distribution, calculated from the mean
                # and fit error bar.
                if self.tree_settings['Stats/plot_theory']:
                    
                    x0  = _n.average(self.plot_parameters[2*n+2]) 
                    ex  = self.plot_parameters[2*n+3][-1]
                    x   = _n.linspace(x0-4*ex, x0+4*ex, 400)
                    pdf = len(self.plot_parameters[1]) * _stats.norm.pdf((x-x0)/ex)/ex * (B[1]-B[0])
                    
                    self.axes_histograms[n+2].plot(x,pdf,label='Expected')
                    self.axes_histograms[n+2].legend()
                
        
        self.figure_stats.canvas.draw()
        self.window.process_events()
        
    def update_all_plots(self, *a):
        """
        Updates the Fit and Stats plots.
        """
        self.update_fit_plot()
        self.update_histograms_plot()
    
    def event_close(self, *a):
        """
        Quits acquisition when the window closes.
        """
        self.button_loop.set_checked(False)
        
        
if __name__ == '__main__':
    import os as _os
    import shutil as _sh
    if _os.path.exists('egg_settings'): _sh.rmtree('egg_settings')

    
    self = fitting_statistics_demo()


import spinmob as _s
import time    as _t
import struct  as _struct
import numpy   as _n
import os      as _os




def load_chn(path=None, **kwargs):
    """
    Loads a Maestro CHN file at the specified path. Will return a databox with all 
    the information.
    
    Parameters
    ----------
    path=None
        If None, a dialog will pop up. Otherwise specify a valid path (string).
        
    Optional keyword arguments, e.g., delimeter=',', are sent to spinmob.data.databox()
    """
    if path==None: path = _s.dialogs.load(filters="*.Chn")
    if path==None: return
    
    # Create a databox
    d = _s.data.databox(**kwargs)
    
    # Open the file
    fh = open(path,mode="rb")
    
    # Read the buffer
    buffer = fh.read()
    
    # Unpack
    [type, unitnumber, segment_number] = _struct.unpack("hhh",buffer[0:6])
    ascii_seconds = buffer[6:8].decode()
    [real_time_20ms, live_time_20ms] = _struct.unpack("ii",buffer[8:16])
    start_date = buffer[16:24].decode()
    start_time = buffer[24:28].decode()
    [channel_offset, channel_count] = _struct.unpack("hh",buffer[28:32])
    
    # Get the counts data
    spectrum = _n.zeros(channel_count,dtype=int)
    for i in range(channel_count): [spectrum[i]] = _struct.unpack("I",buffer[32+4*i:36+4*i])
    d['Channel'] = range(channel_count)
    d['Counts']  = spectrum
    
    # Get the date in a nice format
    if "1" == start_date[7]: century = 20
    else:                    century = 19
    start_RFC2822 = "%s %s %02d%s %s:%s:%s" % (start_date[0:2], start_date[2:5], century, start_date[5:7], start_time[0:2], start_time[2:4], ascii_seconds)
    
    # Header info
    d.path=path
    d.h(path=path)
    d.h(start_time = _t.strptime(start_RFC2822,"%d %b %Y %H:%M:%S"))
    d.h(real_time = 0.02*real_time_20ms)
    d.h(live_time = 0.02*live_time_20ms)
    
    return d

def load_chns(paths=None, **kwargs):
    """
    Loads multiple chn files, returning a list of databoxes.
    
    Parameters
    ----------
    paths=None
        If None, a dialog will pop up, allowing you to select multiople files.
        Otherwise, specify a valid *list* of paths, e.g. ['C:/test/path.chn', 'C:/test/path2.chn']
    
    Optional keyword arguments (e.g., delimiter=',') are sent to load_chn()
    """
    if paths==None: paths=_s.dialogs.load_multiple(filters='*.Chn', **kwargs)
    if paths==None: return
    
    ds = []
    for path in paths: ds.append(load_chn(path, **kwargs))
    return ds

def plot_chn_files(xscript='d[0]', yscript='d[1]', eyscript='sqrt(d[1])', marker='+', linestyle='', xlabel='Channel', ylabel='Counts', paths=None, **kwargs):
    """
    Opens a bunch of chn files and plots the specified script for each. You can
    get the same result by loading multiple chn files (returns a list of databoxes)
    and sending the result to spinmob.plot.xy.databoxes().
    
    Arguments
    ---------
    xscript, yscript, eyscript
        Scripts for creating the xdata, ydata, and error bars. These follow the
        convention of spinmob.plot.xy.files(). Any valid python string evaluating to
        an array of numbers will work. "d" refers to the databox being plotted, and
        the system "knows" about all numpy functions. Specifying None for xscript
        or yscript will generate integer arrays to match the other data sets.
    
    marker, linestyle, xlabel, ylabel
        Some common plot options.
    
    Additional optional keyword arguments are sent to spinmob.plot.databoxes,
    spinmob.plot.data, and pylab.errorbar.
    """
    if paths==None: paths = _s.dialogs.load_multiple(filters='*.Chn')
    if paths==None: return
    
    # Load the files
    ds = load_chns(paths)
    
    # Get the title
    title = _os.path.split(ds[0].path)[0]
    
    _s.plot.xy.databoxes(ds, xscript, yscript, eyscript, 
                         marker=marker, linestyle=linestyle, 
                         xlabel=xlabel, ylabel=ylabel, 
                         title=title, **kwargs)
    
def convert_chn_to_csv(chn_paths=None, output_dir=None):
    """
    Opens the supplied Maestro Chn files and saves them as csv

    Parameters
    ----------
    chn_paths=None : list of strings
        List of paths to Chn files. If None, this will pop up a dialog.
    
    output_dir=None: string
        Path to output directory. If None, this will pop up a dialog.
    """
    ds = load_chns(chn_paths, delimiter=',')
    if output_dir is None: output_dir = _s.dialogs.select_directory('Select an output directory.')
    
    for d in ds: 
        
        # Get the file name
        filename = _os.path.split(d.path)[-1]
        
        # Replace the extension
        s = filename.split('.')
        s[-1] = '.csv'
        filename = '.'.join(s)
        
        # Save it
        d.save_file(_os.path.join(output_dir, filename))
    
    return ds


# Only import this if imageio is installed
try:
    import imageio as _imageio
    
    def load_image(path=None):
        """
        Loads an image, returning a 3D array: two indices for the pixel position
        (x and y), and one index for the color, which can have either 3 values
        (RGB) or 4 (RGBA), depending on the file format.
        
        Parameters
        ----------
        path=None
            If None, a dialog will open up. Otherwise, specify a valid path (string).
        
        """
        if path == None: path = _s.dialogs.load()
        if path == None: return
        
        # Load the image into array "a". 
        return _imageio.imread(path)
        
    
    def load_images(paths=None):
        """
        Loads multiple images, returning a list of 3D arrays, as per load_image.
        
        Parameters
        ----------
        paths=None
            If None, a dialog will pop up. Otherwise, specify a *list* of valid
            paths to images.
        """
        if paths == None: paths = _s.dialogs.load_multiple()
        if paths == None: return
        
        arrays = []
        for path in paths: arrays.append(load_image(path))
        return arrays

except:
    _installimageio = "Note: Image loading is disabled. You may need to install imageio, which can be done with the command 'conda install imageio' from the system terminal or anaconda prompt. You can also find it in anaconda-navigator."
    print(_installimageio)
    def load_image (path =None): print(_installimageio)
    def load_images(paths=None): print(_installimageio) 
    
    


V1 = d['V1X']+1j*d['V1Y']
V2 = d['V2X']+1j*d['V2Y']

x = ( d['f'] )
y = ( abs(V1), angle(V1)*180/pi, abs(V2), angle(V2)*180/pi )

xlabels = 'Frequency (Hz)'
ylabels = ('V1 Mag (V)', 'V1 Phase (deg)', 'V2 Mag (V)', 'V2 Phase (deg)' )


import mcphysics as _mp; libm2k = _mp._m2k

# import libm2k
import pylab
import numpy

# Get the first m2k name
name = libm2k.getAllContexts()[0]

# Get the m2k
m2k = libm2k.m2kOpen(name).toM2k()

# Run the calibration
m2k.calibrate()

# Enable the first output, enable it, and send it some data.
ao = m2k.getAnalogOut()
ao.enableChannel(0, True)
ao.enableChannel(1, True)
ao.setSampleRate(0, 75000000)
ao.setSampleRate(1, 75000000)

# Waveforms: sinusoid and square pulse
N  = 100                        # Number of samples to stream
ns = numpy.linspace(0,N-1,N)    # List of integers
v1 = numpy.sin(2*numpy.pi*ns/N) # Single cycle in N points
v2 = numpy.linspace(1,0,N)      # Trigger edge
v2[0] = 0.5                     # But a little smoother...

# Send both waveforms at once
ao.push([v1,v2]) # Default is cyclic, pulse on the first output

# Enable the first input channel, enable it, and take data
ai = m2k.getAnalogIn()
ai.enableChannel(0, True)
ai.enableChannel(1, True)
ai.setSampleRate(100000000)
ai.setRange(0, libm2k.HIGH_GAIN)
ai.setRange(1, libm2k.HIGH_GAIN)

# Now try triggering
trig = ai.getTrigger()
trig.setAnalogMode(1,1)         # ch2, analog condition
trig.setAnalogSource(1)         # ch2
trig.setAnalogCondition(1,0)    # ch2, rising edge
trig.setAnalogLevel(0,0.5)      # ch2, 0.5V

# Set the timeout (5 sec) to be safe if it doesn't trigger
m2k.setTimeout(3000)

# Go for it
for n in range(10):
    y1, y2 = ai.getSamples(N*3)
    pylab.plot(y1)
    pylab.plot(y2)
